/*
 * tag: dict management
 * 
 * Copyright (C) 2003 Patrick Mauritz, Stefan Reinauer
 *
 * See the file "COPYING" for further information about
 * the copyright and warranty status of this work.
 */

#include <stdlib.h>
#include <string.h>

#include "config.h"
#include "types.h"
#include "dict.h"

void printk(const char *format, ...);

unsigned char *dict = 0;
ucell *last;
cell dicthead = 0;

/* lfa2nfa
 * converts a link field address to a name field address,
 * i.e find pointer to a given words name
 */

ucell lfa2nfa(ucell ilfa)
{
	ilfa = ilfa - (cell) dict;
	ilfa--;			/* skip status */
	while (dict[--ilfa] == 0);
	ilfa -= (dict[ilfa] - 128);
	return ilfa + (cell) dict;
}

/* lfa2cfa
 * converts a link field address to a code field address.
 * in this forth implementation this is just a fixed offset 
 */

static ucell lfa2cfa(ucell ilfa)
{
	return ilfa + sizeof(cell);
}


/* fstrlen - returns length of a forth string. */

static ucell fstrlen(ucell fstr)
{
	fstr -= (ucell) dict;
	while (dict[++fstr] < 128);
	return dict[fstr] - 128;
}

/* tolower - convert a character to lowecase */

static int tolower(int c)
{
	return ((c >= 'A') && (c <= 'Z')) ? (c - 'A' + 'a') : c;
}

/* fstrcmp - compare null terminated string with forth string. */

static int fstrcmp(char *s1, ucell fstr)
{
	while (*s1) {
		if (tolower(*(s1++)) != tolower(*(((char *) fstr)++)))
			return -1;
	}

	return 0;
}

/* findword
 * looks up a given word in the dictionary. This function
 * is used by the c based interpreter and to find the "initialize"
 * word.
 */

ucell findword(char *s1)
{
	ucell tmplfa, len;

	if (!last)
		return (ucell) - 1;

	tmplfa = *last;

	len = strlen(s1);

	while (tmplfa) {
		ucell nfa = lfa2nfa(tmplfa);

		if (len == fstrlen(nfa) && !fstrcmp(s1, nfa))
			return lfa2cfa(tmplfa);

		tmplfa = *(ucell *) tmplfa;
	}

	return (ucell) - 1;
}


ucell load_dictionary(char *data, ucell len)
{
	ucell *walk, head, intlast;

	/* assertions */
	if (len <= (16 + 2 * sizeof(cell)) || strncmp(DICTID, data, 16))
		return 0;

	data += 16;
	len -= 16;

	dicthead = len - 2 * sizeof(cell);
	memcpy(dict, data, dicthead);

	head = *(cell *) (data + dicthead);

	intlast = (*(cell *) (data + dicthead + sizeof(cell)));

	intlast += (cell) dict - head;

#ifdef DEBUG_DICTIONARY
	printk("\ndynamic relocation...");
#endif
	for (walk = (ucell *) dict; walk < (ucell *) (dict + dicthead);
	     walk++) {
		if (*walk > head && *walk < head + dicthead)
			*walk = (ucell) (*walk - head + dict);
	}

#ifdef DEBUG_DICTIONARY
	printk(" done.\n");
#endif

	last = (ucell *) intlast;

	return -1;
}
