/* tag: internal words, inner interpreter and such
 *
 * Copyright (C) 2003 Patrick Mauritz, Stefan Reinauer
 *
 * See the file "COPYING" for further information about
 * the copyright and warranty status of this work.
 */

/*
 * execution works as follows:
 *  - PC is pushed on return stack
 *  - PC is set to new CFA
 *  - address pointed by CFA is executed by CPU
 */

extern void *words[];
ucell PC;
volatile int runforth=0;

static inline void processxt(ucell xt)
{
	void (*tokenp) (void);
#ifdef DEBUG_INTERPRETER
	printk("processxt: pc=%x, xt=%x\n", PC, xt);
#endif
	tokenp = words[xt];
	tokenp();
}

static void docol(void)
{				/* DOCOL */
	PUSHR(PC);
	PC = *(ucell *) PC;
#ifdef DEBUG_INTERPRETER
	printk("docol: %s\n", lfa2nfa(PC - sizeof(ucell)));
#endif
}

static void semis(void)
{
	PC = POPR();
}

static inline void next(void)
{
	PC += sizeof(ucell);
#ifdef DEBUG_INTERPRETER
	printk("next: PC is now %x\n", PC);
#endif
	processxt(*(ucell *) (*(ucell *) PC));
}

void enterforth(ucell _cfa)
{
	ucell tmp;

	if (*(ucell *) _cfa == DOCOL) {
#ifdef DEBUG_INTERPRETER
		printk("enterforth: COLON started -- rstackcnt: %x\n",
		       rstackcnt);
#endif
		if (rstackcnt < 0) {
			rstackcnt = 0;
		}
		tmp = rstackcnt;
		
		runforth=1;
		
		PUSHR(PC);
		PC = _cfa;
		while (rstackcnt > tmp && runforth) {
#ifdef DEBUG_INTERPRETER
			printk("enterforth: NEXT\n");
#endif
			next();
		}
		POPR();
	} else {
#ifdef DEBUG_INTERPRETER
		printk("not COLON started\n");
#endif
		tmp = _cfa;
		PC = (ucell) & tmp;
#ifdef DEBUG_INTERPRETER
		printk("PC=%x, tmp=%x, cfa=%x\n", PC, tmp, _cfa);
#endif
		processxt(*(ucell *) _cfa);
	}
}

/* called inline thus a slightly different behaviour */
static void lit(void)
{				/* LIT */
	PC += sizeof(cell);
	PUSH(*(ucell *)PC);
#ifdef DEBUG_INTERPRETER
	printk("lit: %x\n", *(ucell *)PC);
#endif
}

static void docon(void)
{				/* DOCON */
	ucell tmp = *(ucell *) (*(ucell *) PC + sizeof(ucell));
	PUSH(tmp);
#ifdef DEBUG_INTERPRETER
	printk("docon: PC=%x, value=%x\n", PC, tmp);
#endif
}

static void dovar(void)
{				/* DOVAR */
	ucell tmp = (*(ucell *) PC) + sizeof(ucell);
	PUSH(tmp);		/* returns address to variable */
#ifdef DEBUG_INTERPRETER
	printk("dovar: PC: %x, %x\n", PC, tmp);
#endif
}

static void dobranch(void)
{				/* unconditional branch */
	PC += sizeof(cell);
	PC += *(cell *) PC;
}

static void docbranch(void)
{				/* conditional branch */

	PC += sizeof(cell);
	if (POP()) {
#ifdef DEBUG_INTERNAL
		printk("  ?branch: end loop\n");
#endif
	} else {
#ifdef DEBUG_INTERNAL
		printk("  ?branch: follow branch\n");
#endif
		PC += *(cell *) PC;
	}
}

static ucell trampoline_semis = 0;
static ucell trampoline[3] = { DOCOL, 0, (ucell) & trampoline_semis };

static void execute(void)
{				/* EXECUTE */
	ucell address = POP();
#ifdef DEBUG_INTERPRETER
	printk("execute: %x\n", address);
#endif
	PUSHR(PC);
	trampoline[1] = address;
	PC = (ucell) & trampoline;
}

static void dodoes(void)
{				/* DODOES */
	ucell data = *(ucell *) PC + (2 * sizeof(ucell));
	ucell word = *(ucell *) (*(ucell *) PC + sizeof(ucell));

#ifdef DEBUG_INTERPRETER
	printk("DODOES data=%x word=%x\n", data, word);
#endif

	PUSH(data);
	PUSH(word);

	execute();
}

static void dodefer(void)
{
	docol();
}

static void dodo(void)
{
	cell startval, endval;
	startval = POP();
	endval = POP();

	PUSHR(endval);
	PUSHR(startval);
}

static void doisdo(void)
{
	cell startval, endval, offset;

	startval = POP();
	endval = POP();

	PC += sizeof(cell);

	if (startval == endval) {
		offset = *(cell *) PC;
		PC += offset;
	} else {
		PUSHR(endval);
		PUSHR(startval);
	}
}

static void doloop(void)
{
	cell offset, startval, endval;

	startval = POPR() + 1;
	endval = POPR();

	PC += sizeof(cell);

	if (startval < endval) {
		offset = *(cell *) PC;
		PC += offset;
		PUSHR(endval);
		PUSHR(startval);
	}

}

static void doplusloop(void)
{
	ucell high, low;
	cell increment, startval, endval, offset;

	increment = POP();

	startval = POPR();
	endval = POPR();

	low = (ucell) startval;
	startval += increment;

	PC += sizeof(cell);

	if (increment >= 0) {
		high = (ucell) startval;
	} else {
		high = low;
		low = (ucell) startval;
	}

	if (endval - (low + 1) >= high - low) {
		offset = *(cell *) PC;
		PC += offset;

		PUSHR(endval);
		PUSHR(startval);
	}
}
