/* tag: openbios forth environment, executable code
 *
 * Copyright (C) 2003 Patrick Mauritz, Stefan Reinauer
 *
 * See the file "COPYING" for further information about
 * the copyright and warranty status of this work.
 */

#include "config.h"
#include "types.h"
#include "dict.h"
#define NATIVE
#include "stack.h"

void enterforth(ucell);
void printk(const char *fmt, ...);
void cls(void);
#ifdef DEBUG_CONSOLE
int uart_init(int port, unsigned long speed);
#endif

extern unsigned char *dict;
extern cell dicthead;

static char intdict[64 * 1024];

void init_memory(void)
{
	/* push start and end of available memory to the stack
	 * so that the forth word QUIT can initialize memory 
	 * management. For now we use hardcoded memory between
	 * 0x10000 and 0x9ffff (576k). If we need more memory
	 * than that we have serious bloat.
	 */

	PUSH(0x10000);
	PUSH(0x9FFFF);
}

/* Check if MAGIC is valid and print the Multiboot information structure
   pointed by ADDR.  */

void openbios(ucell dictstart, ucell dictend)
{
#ifdef DEBUG_CONSOLE
	uart_init(SERIAL_PORT, SERIAL_SPEED);
	/* Clear the screen.  */
	cls();
#endif

	dict = intdict;

#ifdef DEBUG_BOOT
	printk("forth started.\n");
	printk("initializing memory...");
#endif

	init_memory();

#ifdef DEBUG_BOOT
	printk("done\n");
	printk("loading dictionary... ");
#endif

#ifdef DEBUG_DICTIONARY
	printk("dictionary start: 0x%x\n", dictstart);
	printk("dictionary end  : 0x%x\n", dictend);
#endif
	if (!dictstart || !dictend) {
		printk("panic: no dictionary.\n");
		return;
	}

	load_dictionary((char *) dictstart, dictend - dictstart);

	PC = findword("initialize");

	if (PC == (ucell) (-1)) {
		printk("panic: dictionary not valid\n");
		return;
	}
#ifdef DEBUG_DICTIONARY
	printk("done (%d bytes).\n", dicthead);
	printk("Jumping to dictionary...\n");
#endif

	enterforth(PC);

	return;
}
