/* tag: openbios qt plugin skeleton
 * 
 * Copyright (C) 2003 Stefan Reinauer
 *
 * See the file "COPYING" for further information about
 * the copyright and warranty status of this work.
 */


extern "C" {
#include <pthread.h>
#include <unistd.h>
#include <plugins.h>
#include <plugin_pci.h>
}
#include "plugin_qt.h"
#include "pciconfig.h"

#define DEBUG

volatile unsigned char * fb=0;
volatile int gui_running=0;

typedef struct {
	int argc;
	char **argv;
} threaddata;

void *gui_thread(void *ptr)
{
	threaddata *td=(threaddata *)ptr;
	
	QApplication a(td->argc, td->argv);
	FrameBufferWidget w;
	
	a.setMainWidget(&w);
	w.show();
	
	fb=w.getFrameBuffer();

	gui_running=-1;
	a.exec();
	gui_running=0;
	
	return 0;
}

extern "C" {
int plugin_qt_init(void)
{
	pthread_t mythread;
	char *args[]={ "plugin_qt" };
	threaddata mytd = { 1, args };
#ifdef DEBUG
	printf("Initializing \"framebuffer\" plugin...");
#endif
	pthread_create(&mythread, NULL, gui_thread, &mytd);
	while (!fb) 
		usleep(20);

	/* now we have the framebuffer start address.
	 * updating pci config space to reflect this
	 */
	
	/* uarghs.. this will fail on 64bit architectures
	 * FIXME FIXME FIXME FIXME 
	 */
	*(u32 *)(pci_config_space+0x10)=(u32)fb;
	
	/* FIXME: we need to put the fcode image for this
	 * device to the rom resource, once it exists
	 */
	
	/* register pci device to be available to beginagain */
	pci_register_device(0, 2, 0, pci_config_space);
	
#ifdef DEBUG
	printf("done.\n");
#endif
	return 0;
}
PLUGIN_AUTHOR("Stefan Reinauer <stepan@openbios.org>")
PLUGIN_DESCRIPTION("QT gui plugin emulating framebuffer device")
PLUGIN_LICENSE("GPL v2")
PLUGIN_DEPENDENCIES("pci")

}

#if 0
int main (int argc, char *argv[])
{
	init_plugin(argc, argv);

#if 1
	/* demo writes */
	for (int i=0; i< 640; i++)
		for (int j=0; j<480; j++)
			fb[i+j*640]=15;

        for ( int i = 0; i < 400; i++ )
		fb[i*640 + i] = i % 255;
#endif
	
	while (gui_running)
	{
		sleep(1);
	}
	printf("gui ended regularly.\n");
}
#endif
